from __future__ import annotations

__all__ = [
    "JSONArray",
    "JSONObject",
    "JSON",
    "DecimalEncoder",
    "is_json_str",
    "load_json",
    "load_json_object",
]

import decimal
import json
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Protocol, Type, TypeVar, Union, cast

# TODO: Figure out better JSON typing. mypy doesn't like
JSON = Union[Dict[str, Any], List[Any], int, str, float, bool, Type[None]]
# JSON = Union['JSONArray', 'JSONObject', int, str, float, bool, Type[None]]


if TYPE_CHECKING:  # pragma: no cover

    class JSONArray(list[JSON], Protocol):  # type: ignore
        # __class__: Type[list[JSON]]  # type: ignore[assignment]
        pass

    class JSONObject(dict[str, JSON], Protocol):  # type: ignore
        # __class__: Type[dict[str, JSON]]  # type: ignore[assignment]
        pass

else:
    JSONArray, JSONObject = List[Any], Dict[str, Any]


class DecimalEncoder(json.JSONEncoder):
    """Used to encode decimal.Decimal when printing/encoding dicts to
    JSON strings
    """

    def default(self, o: Any) -> Union[str, json.JSONEncoder]:
        if isinstance(o, decimal.Decimal):
            return str(o)
        return super(DecimalEncoder, self).default(o)


def is_json_str(data: Any) -> bool:
    try:
        assert isinstance(data, str)
        json.loads(data)
    except:
        return False
    return True


def load_json(path_or_str: Union[str, Path], **kwargs) -> JSON:
    if isinstance(path_or_str, str) and is_json_str(path_or_str):
        return json.loads(path_or_str, **kwargs)
    elif Path(path_or_str).exists():
        with open(str(path_or_str), "r") as f:
            return json.load(f, **kwargs)
    else:
        raise ValueError(f"Cannot load {path_or_str} as json. Not valid json string or path.")


def load_json_object(path_or_str: Union[str, Path], **kwargs) -> JSONObject:
    json_data = load_json(path_or_str, **kwargs)
    if not isinstance(json_data, dict):
        raise ValueError(f"{path_or_str} was loaded as JSON but not a JSON Object")
    return cast(JSONObject, json_data)



__all__ = [
    "DBModel",
    "DBKeyNameEnum",
    "DBSortKeyNameEnum",
    "DBIndexNameEnum",
    "DBIndex",
]

from dataclasses import dataclass
from typing import List, Optional, Tuple, Type, TypeVar

from aibs_informatics_core.collections import StrEnum
from aibs_informatics_core.env import EnvBase
from aibs_informatics_core.models.base import SchemaModel
from aibs_informatics_core.models.db.type_defs import DynamoDBItemValue, DynamoDBKey


# ========================== DB entry base model =========================
@dataclass
class DBModel(SchemaModel):
    """Base class that DB Entry models should all inherit from.
    Used to disambiguate from other model classes that also inherit from
    SchemaModel (like Request/Response models)"""

    pass


# =========================== Base definitions ===========================
class DBKeyNameEnum(StrEnum):
    """An Enum that describes possible Partition Keys for a DynamoDB table

    NOTE: Convention for this Enum should be that the first enum member is the
          partition key for the main table and subsequent members describe
          partition keys for Global/Local Secondary Indices (GSIs/LSIs)
    """

    pass


class DBSortKeyNameEnum(StrEnum):
    """An Enum that describes possible Sort Keys for a DynamoDB table.

    NOTE: Not all tables/GSI's/LSI's will have a sort key
    """

    pass


class DBIndexNameEnum(StrEnum):
    """An Enum that describes names for Global/Local Secondary Indices for
    a given table.

    NOTE: Not all tables will have GSIs/LSIs
    """

    @staticmethod
    def from_name_and_key(
        table_name: str, key: DBKeyNameEnum, sort_key: Optional[DBSortKeyNameEnum] = None
    ) -> str:
        """Construct an AWS DynamoDB Secondary Index name from table_name, key, and sort_key.

        NOTE: `key` and `sort_key` names will have underscores ('_') converted to hyphens ('-')
        """
        index_name = f"{table_name}-{key.value.replace('_', '-')}"
        if sort_key:
            index_name = f"{index_name}-{sort_key.value.replace('_', '-')}"
        return f"{index_name}-index"


DB_INDEX = TypeVar("DB_INDEX", bound="DBIndex")


class DBIndex(StrEnum):
    """Helper Enum class that indexes partition keys, sort keys, and possible GSIs/LSIs

    Classes inheriting from the DBIndex should:
    - Provide an implementation of the `table_name` classmethod
    - Starting with the main table enum member, define a tuple consisting of:
        1. canonical partition key value
        2. DBKeyNameEnum representing the partition key
        3. DBSortKeyNameEnum representing the sort key (or None if table/LSI/GSI lacks a sort key)
        4. DBIndexNameEnum representing the name of the LSI/GSI if it exists (or None)
           NOTE: main table is not an LSI/GSI so this item will always be None for those.
        5. Optional list of strings representing the non-key attributes for a GSI
           NOTE: this should only be used for GSIs
    """

    _value_: str
    _key_name: DBKeyNameEnum
    _sort_key_name: Optional[DBSortKeyNameEnum]
    _index_name: Optional[DBIndexNameEnum]
    _attributes: Optional[List[str]]
    _all_values: Tuple[
        str,  # _value_
        DBKeyNameEnum,  # _key_name
        Optional[DBSortKeyNameEnum],  # _sort_key_name
        Optional[DBIndexNameEnum],  # _index_name
        Optional[List[str]],  # _attributes
    ]

    def __new__(cls, *values):
        obj = str.__new__(cls, values[0])
        # first value is canonical value
        obj._value_ = values[0]
        obj._key_name = values[1]
        # cls._value2member_map_[obj._key_name] = obj
        obj._sort_key_name = values[2]
        obj._index_name = values[3]
        obj._attributes = values[4] if len(values) > 4 else None
        obj._all_values = (
            obj._value_,
            obj._key_name,
            obj._sort_key_name,
            obj._index_name,
            obj._attributes,
        )
        return obj

    @classmethod
    def table_name(cls) -> str:
        raise NotImplementedError(  # pragma: no cover
            f"Enum inheriting from DBIndex ({cls.__name__}) needs to implement "
            "the table_name classmethod!"
        )

    @property
    def supports_strongly_consistent_read(self) -> bool:
        # Only the main DynamoDB table supports strongly consistent reads
        # See: https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/HowItWorks.ReadConsistency.html
        return self._index_name is None

    @property
    def key_name(self) -> str:
        return self._key_name.value

    @property
    def sort_key_name(self) -> Optional[str]:
        if self._sort_key_name is not None:
            return self._sort_key_name.value
        return None

    @property
    def index_name(self) -> Optional[str]:
        if self._index_name is not None:
            return self._index_name.value
        return None

    @property
    def non_key_attributes(self) -> Optional[List[str]]:
        return self._attributes

    def is_partial(self) -> bool:
        return self._attributes is not None

    def get_key_filter(
        self,
        partition_value: DynamoDBItemValue,
        sort_value: Optional[DynamoDBItemValue] = None,
    ) -> DynamoDBKey:
        """Create a DynamoDB key filter for primary or composite (partition + sort) keys

        For composite keys it is okay to provide only a 'partition value' without the 'sort_value'

        Args:
            partition_value (DynamoDBItemValue): Partition key value
            sort_value (Optional[DynamoDBItemValue], optional): Optional sort key value. Defaults to None.

        Returns:
            DynamoDBKey: A DynamoDBKey mapping that maps partition and sort key names to values
        """
        return self.get_primary_key(
            partition_value=partition_value, sort_value=sort_value, strict=False
        )

    def get_primary_key(
        self,
        partition_value: DynamoDBItemValue,
        sort_value: Optional[DynamoDBItemValue] = None,
        strict: bool = True,
    ) -> DynamoDBKey:
        """Create a DynamoDB primary (primary or partition + sort) key for read/write Table item operations.

        This method needs to be strict about always providing a `sort_value` if a table/GSI/LSI
        has a 'sort key' because its output is intended to be used with functions like
        Boto3 DynamoDB Table get-item
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/table/get_item.html#get-item
        where both the partition and sort components of primary key *must* be provided.

        Args:
            partition_value (DynamoDBItemValue): Partition key value
            sort_value (Optional[DynamoDBItemValue], optional): Optional sort key value. Defaults to None.
            strict (bool): Whether a non-default 'sort_value' MUST be provided if a 'sort key'
                exists. Defaults to True ('sort_value' MUST be provided if 'sort key' exists).

        Raises:
            ValueError: If 'sort key' is undefined, but a sort key value IS provided
            ValueError: If strict=True AND if 'sort key' is defined, but NO sort key value is provided

        Returns:
            DynamoDBKey: A DynamoDBKey mapping that maps partition and sort key names to values
        """
        if self.key_name is None:
            raise ValueError(f"'{self}' has no key name.")
        db_key = {self.key_name: partition_value}

        if self.sort_key_name:
            if sort_value is None and strict:
                raise ValueError(
                    f"{self} has a sort key ({self.sort_key_name}) "
                    "but sort key value is unspecified!"
                )

            if sort_value is not None:
                db_key[self.sort_key_name] = sort_value
        else:
            if sort_value is not None:
                raise ValueError(
                    f"'{self}' has no sort key, but a sort key value was given ({str(sort_value)})"
                )
        return db_key

    def get_table_name(self, env_base: EnvBase) -> str:
        return env_base.get_resource_name(self.table_name())

    def get_index_name(self, env_base: EnvBase) -> Optional[str]:
        index_name = self.index_name
        return index_name if index_name is None else env_base.prefixed(index_name)

    @classmethod
    def options(cls) -> List[str]:
        return [_.value for _ in cls]

    @classmethod
    def get_default_index(cls: Type[DB_INDEX]) -> DB_INDEX:
        indexes: List[DB_INDEX] = list(cls)
        if len(indexes) == 0:
            raise ValueError(f"{cls.__name__} has no members!")

        for index in indexes:
            # primary index is the first index in the list and does not have name
            if index.index_name is None and not index.is_partial():
                return index
        else:
            raise ValueError(f"{cls.__name__} has no primary index!")



__all__ = [
    "DynamoDBItemKey",
    "DynamoDBItemValue",
    "DynamoDBKey",
    "DynamoDBPrimaryKeyItemValue",
]

from decimal import Decimal
from typing import Any, Mapping, Sequence, Set, Union

DynamoDBPrimaryKeyItemValue = Union[
    bytes,
    str,
    int,
    Decimal,
    bool,
]

# This alias refers to the allowed types as defined in `mypy_boto3_dynamodb`
# https://youtype.github.io/boto3_stubs_docs/mypy_boto3_dynamodb/service_resource/#tableget_item-method
DynamoDBItemValue = Union[
    bytes,
    bytearray,
    str,
    int,
    Decimal,
    bool,
    Set[int],
    Set[Decimal],
    Set[str],
    Set[bytes],
    Set[bytearray],
    Sequence[Any],
    Mapping[str, Any],
    None,
]
DynamoDBItemKey = str
DynamoDBKey = Mapping[DynamoDBItemKey, DynamoDBItemValue]



import uuid
from typing import Any, ClassVar, List, Optional, Type, TypeVar, Union

import marshmallow as mm

from aibs_informatics_core.models.base import CustomStringField
from aibs_informatics_core.utils.hashing import uuid_str
from aibs_informatics_core.utils.os_operations import get_env_var

UNIQUE_ID_TYPE = TypeVar("UNIQUE_ID_TYPE", bound="UniqueID")


class UniqueID(str):
    """An augmented `str` class intended to represent a unique ID type"""

    ENV_VARS: ClassVar[List[str]] = ["UNIQUE_ID"]

    def __new__(cls, *args, **kwargs):
        return str.__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        try:
            uuid_obj = uuid.UUID(self, version=4)
        except ValueError:
            raise mm.ValidationError(f"'{self}' is not a valid {self.__class__.__name__} (uuid4)!")
        self._uuid_obj = uuid_obj

    @classmethod
    def as_mm_field(cls, *args, **kwargs) -> mm.fields.Field:
        return CustomStringField(cls, *args, **kwargs)

    @classmethod
    def create(
        cls: Type[UNIQUE_ID_TYPE], seed: Optional[Union[int, str]] = None
    ) -> UNIQUE_ID_TYPE:
        return cls(uuid_str(str(seed)) if seed is not None else uuid.uuid4())

    def as_uuid(self) -> uuid.UUID:
        return self._uuid_obj

    @classmethod
    def from_env(cls: Type[UNIQUE_ID_TYPE]) -> UNIQUE_ID_TYPE:
        env_var = get_env_var(*cls.ENV_VARS)
        if env_var is None:
            raise ValueError(
                f"Could not find environment variable for {cls} given ENV VARS: {cls.ENV_VARS}"
            )
        return cls(env_var)



import logging
import os
from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union, cast

import marshmallow as mm

from aibs_informatics_core.exceptions import ValidationError
from aibs_informatics_core.models.aws.s3 import S3URI
from aibs_informatics_core.models.base import (
    DictField,
    ListField,
    SchemaModel,
    StringField,
    custom_field,
)
from aibs_informatics_core.models.base.custom_fields import UnionField
from aibs_informatics_core.models.demand_execution.job_param import (
    DownloadableJobParam,
    JobParam,
    JobParamEnvName,
    JobParamRef,
    ResolvableJobParam,
    UploadableJobParam,
)
from aibs_informatics_core.models.demand_execution.job_param_resolver import JobParamResolver
from aibs_informatics_core.models.demand_execution.param_pair import (
    JobParamPair,
    JobParamSetPair,
    ParamPair,
    ParamSetPair,
)
from aibs_informatics_core.models.demand_execution.resolvables import (
    Resolvable,
    StringifiedUploadable,
    Uploadable,
    get_resolvable_from_value,
)
from aibs_informatics_core.utils.hashing import sha256_hexdigest
from aibs_informatics_core.utils.json import JSON

logger = logging.getLogger(__name__)


class refresh_params:
    def __init__(
        self, func: Optional[Callable] = None, force: bool = True, pre_validate: bool = False
    ):
        self.force = force
        self.pre_validate = pre_validate
        self.func = func

    def __get__(self, instance, owner):
        return partial(self.__call__, instance)

    def __call__(self, *args, **kwargs):
        # If function has not been set, the __call__ is passing in the function
        # Result: we set self.func and return self
        if not self.func:
            self.func = cast(Callable, args[0])
            return self
        # Otherwise, we call refresh once the already set function has finished
        obj = cast(DemandExecutionParameters, args[0])
        if self.pre_validate:
            obj._refresh(force=False)
        value = self.func(*args, **kwargs)
        obj._refresh(force=self.force)
        return value


@dataclass
class DemandExecutionParameters(SchemaModel):
    command: List[str] = custom_field(default_factory=list, mm_field=ListField(StringField))
    params: Dict[str, Any] = custom_field(default_factory=dict, mm_field=DictField(StringField))
    inputs: List[str] = custom_field(default_factory=list, mm_field=ListField(StringField))
    outputs: List[str] = custom_field(default_factory=list, mm_field=ListField(StringField))
    outputs_metadata: Dict[str, Dict[str, JSON]] = custom_field(
        default_factory=dict, mm_field=DictField(keys=StringField(), values=DictField(StringField))
    )
    output_s3_prefix: Optional[S3URI] = custom_field(default=None, mm_field=S3URI.as_mm_field())
    param_pair_overrides: Optional[List[Union[ParamSetPair, ParamPair]]] = custom_field(
        default=None,
        mm_field=ListField(
            UnionField(
                [(ParamSetPair, ParamSetPair.as_mm_field()), (ParamPair, ParamPair.as_mm_field())]
            )
        ),
    )
    verbosity: bool = custom_field(default=False)

    def __post_init__(self):
        self._refresh_hash = self._compute_refresh_hash()
        self._refresh(True)

    def validate_parameters(
        self, check_input_output_params: bool = True, check_param_pairs: bool = True
    ):
        """Validates Parameters

        Validation Steps:
            * Validate params do not have conflicting envnames
            * Validate inputs/outputs do not overlap
            * Validate that inputs/outputs are found in params
            * TODO: Validate that command only references valid parameters
        """
        # Validate params
        if check_input_output_params:
            self._validate_params()

        # Validate input_output_map
        if check_param_pairs:
            self._validate_param_pairs()

        # TODO: we need to check command

    def _validate_params(self):
        # Validate that input/output job env name parameters do not collide
        inp_envnames = [JobParam.as_envname(_) for _ in self.inputs]
        out_envnames = [JobParam.as_envname(_) for _ in self.outputs]
        envname_intersection = set(inp_envnames).intersection(out_envnames)
        if envname_intersection:
            raise ValidationError(
                f"Job inputs and outputs have overlapping env variable names: "
                f"{envname_intersection}"
            )

        # Validate that inputs/outputs are found in params
        param_map = self.job_param_map
        missing_param_envnames = set(inp_envnames).union(out_envnames).difference(param_map.keys())
        if len(missing_param_envnames) > 0:
            raise ValidationError(
                f"Batch Job inputs/outputs not found in param envnames: "
                f"{missing_param_envnames}"
            )

    def _validate_param_pairs(self):
        # Validate that all param pairs are inputs/outputs
        all_input_output_set = set().union(
            *[_.inputs.union(_.outputs) for _ in self.param_set_pairs]
        )
        diff = all_input_output_set.difference(set(self.inputs).union(set(self.outputs)))
        if len(diff) > 0:
            raise ValidationError(
                "input_output_mapping contained value(s) not found in params: " f"{diff}"
            )

        # Validate no duplicate output sets
        seen = set()
        duplicate_output_sets = []
        for s in self.param_set_pairs:
            if s.outputs in seen:
                duplicate_output_sets.append(s.outputs)
            if s.outputs:
                seen.add(s.outputs)
        if len(duplicate_output_sets) > 0:
            raise ValidationError(
                "Duplicate output set(s) in input_output_map: " f"{duplicate_output_sets}"
            )

    @refresh_params(force=False)
    def add_inputs(self, *param_keys, **param_key_values):
        """Add new input keys to the execution parameters.

        Inputs provided without values are assumed to be in params already
        """
        for i in [*param_keys, *param_key_values.keys()]:
            if i not in self.inputs:
                self.inputs.append(i)
        self.update_params(**param_key_values)

    @refresh_params(force=False)
    def add_outputs(self, *param_keys, **param_key_values):
        """Add new output keys to the execution parameters.

        Outputs provided without values are assumed to be in params already
        """
        for i in [*param_keys, *param_key_values.keys()]:
            if i not in self.outputs:
                self.outputs.append(i)
        self.update_params(**param_key_values)

    @refresh_params(force=False)
    def update_params(self, *param_pairs: Tuple[str, Any], **param_key_values: Any):
        params = dict(param_pairs)
        params.update(param_key_values)
        self.params.update(params)

    def get_param(self, envname: str) -> Optional[Any]:
        """Checks if param contains environment name or placeholder

        Args:
            envname (str): An environment name or placeholder.

        Returns:
            bool: True if param contains envname or placeholder references
        """
        envname = self._sanitize_envname(envname)
        job_param = self._job_param_map.get(envname)
        if job_param:
            return self.params.get(job_param.name)
        return None

    def has_param(self, envname: str) -> bool:
        envname = self._sanitize_envname(envname)
        # Cannot use self.get_param because it will return None if param is not found
        job_param = self._job_param_map.get(envname)
        if job_param:
            return job_param.name in self.params
        return False

    def get_job_param(self, envname: str) -> Optional[JobParam]:
        if JobParamRef.is_valid(envname):
            envname = JobParamRef(envname).envname
        return self.job_param_map.get(JobParam.as_envname(envname))

    def get_input_job_param(self, envname: str) -> Optional[DownloadableJobParam]:
        job_param = self.get_job_param(envname)
        if job_param and isinstance(job_param, DownloadableJobParam):
            return job_param
        return None

    def get_output_job_param(self, envname: str) -> Optional[UploadableJobParam]:
        job_param = self.get_job_param(envname)
        if job_param and isinstance(job_param, UploadableJobParam):
            return job_param
        return None

    @property
    def param_pairs(self) -> List[ParamPair]:
        return [p for ps in self.param_set_pairs for p in ps.to_pairs()]

    @property
    def param_set_pairs(self) -> List[ParamSetPair]:
        param_set_pairs: List[ParamSetPair] = []
        param_pairs: List[ParamPair] = []
        if self.param_pair_overrides:
            seen_outputs: Set[Union[str, None]] = set()
            for pair in self.param_pair_overrides:
                if isinstance(pair, ParamSetPair):
                    seen_outputs.update(pair.outputs)
                    param_set_pairs.append(pair)
                else:
                    seen_outputs.add(pair.output)
                    param_pairs.append(pair)
            if unseen_outputs := set(self.outputs).difference(seen_outputs):
                # NOTE: Here we are allowing for param_pair_overrides to not contain
                #       all inputs/outputs.
                #       We collect remaining
                logger.warning(
                    f"param_pair_overrides did not contain all inputs/outputs: "
                    f"outputs={unseen_outputs}. Adding param set pair to include them."
                )
                param_set_pairs.append(
                    ParamSetPair(frozenset(self.inputs), frozenset(unseen_outputs))
                )
        else:
            param_pairs.extend(ParamPair.from_sets(inputs=self.inputs, outputs=self.outputs))
        param_set_pairs.extend(ParamSetPair.from_pairs(*param_pairs))
        return param_set_pairs

    @property
    def job_param_pairs(self) -> List[JobParamPair]:
        set_pairs: List[JobParamPair] = []
        for pair in self.param_pairs:
            inp_job_param = self.get_input_job_param(pair.input) if pair.input else None
            out_job_param = self.get_output_job_param(pair.output) if pair.output else None
            set_pairs.append(JobParamPair(inp_job_param, out_job_param))
        return set_pairs

    @property
    def job_param_set_pairs(self) -> List[JobParamSetPair]:
        r_job_param_map = {
            k: v for k, v in self.job_param_map.items() if isinstance(v, ResolvableJobParam)
        }
        return [
            JobParamSetPair(
                inputs=frozenset({r_job_param_map[JobParam.as_envname(_)] for _ in pair.inputs}),
                outputs=frozenset({r_job_param_map[JobParam.as_envname(_)] for _ in pair.outputs}),
            )
            for pair in self.param_set_pairs
        ]

    @property
    def resolved_command(self) -> List[str]:
        environment: Dict[str, str] = dict(os.environ)
        environment.update({k: v.value for k, v in self.job_param_map.items()})
        return [
            JobParamRef.replace_references(command_part, environment)
            for command_part in self.command
        ]

    @property
    def job_params(self) -> List[JobParam]:
        return self._job_params

    @property
    def job_param_map(self) -> Dict[str, JobParam]:
        return self._job_param_map

    @property
    def job_param_inputs(self) -> List[JobParam]:
        self._refresh(force=False)
        return [self.job_param_map[JobParam.as_envname(_)] for _ in self.inputs]

    @property
    def job_param_outputs(self) -> List[JobParam]:
        self._refresh(force=False)
        return [self.job_param_map[JobParam.as_envname(_)] for _ in self.outputs]

    @property
    def downloadable_job_param_inputs(self) -> List[DownloadableJobParam]:
        return [_ for _ in self.job_param_inputs if isinstance(_, DownloadableJobParam)]

    @property
    def uploadable_job_param_outputs(self) -> List[UploadableJobParam]:
        return [_ for _ in self.job_param_outputs if isinstance(_, UploadableJobParam)]

    def _sanitize_envname(self, envname: str) -> JobParamEnvName:
        """Converts envname to JobParamEnvName

        This also converts JobParamRef to JobParamEnvName

        Args:
            envname (str): any string reference to parameter

        Returns:
            JobParamEnvName:
        """
        if JobParamRef.is_valid(envname):
            envname = JobParamRef(envname).envname
        return JobParam.as_envname(envname)

    def _param_to_job_params(self) -> List[JobParam]:
        """Convert param dictionary into List of JobParam objects

        Returns:
            List[JobParam]:
        """
        job_params: List[JobParam] = []
        input_envnames = set([JobParam.as_envname(i) for i in self.inputs])
        output_envnames = set([JobParam.as_envname(o) for o in self.outputs])
        for k, v in self.params.items():
            job_param_envname = JobParam.as_envname(k)
            if job_param_envname in input_envnames:
                job_params.append(self._param_to_job_params__build_input(k, v))
            elif job_param_envname in output_envnames:
                job_params.append(self._param_to_job_params__build_output(k, v))
            else:
                job_params.append(JobParam(k, str(v)))
        return job_params

    def _param_to_job_params__build_input(self, k: str, v: Any) -> JobParam:
        resolvable = get_resolvable_from_value(value=v, resolvable_classes=[Resolvable])
        return DownloadableJobParam(k, resolvable.local, resolvable.remote)

    def _param_to_job_params__build_output(self, k: str, v: Any) -> JobParam:
        if StringifiedUploadable.is_valid(v):
            str_uploadable = StringifiedUploadable(v)
            # This logic holds becauses remote is None if destination not found
            if not str_uploadable.remote and not self.output_s3_prefix:
                raise ValueError(
                    f"{str_uploadable} has no destination specified and no output prefix provided"
                )
            remote_value = str_uploadable.remote or S3URI(f"{self.output_s3_prefix}/{v}")
            return UploadableJobParam(k, str_uploadable.local, remote_value)
        else:
            # Only create default remote if value is string and not a stringified uploadable reference
            default_remote = (
                S3URI(f"{self.output_s3_prefix}/{v}")
                if isinstance(v, str) and self.output_s3_prefix
                else None
            )
            uploadable = Uploadable.from_any(value=v, default_remote=default_remote)
            return UploadableJobParam(k, uploadable.local, uploadable.remote)

    def _set_job_params(self, job_params: List[JobParam]):
        self._job_params = JobParamResolver.resolve_references(job_params)
        self._job_param_map = {_.envname: _ for _ in self._job_params}
        self.validate_parameters()

    # ------------------------------------------------
    #                   Refresh methods
    # ------------------------------------------------

    def _refresh(self, force: bool = False):
        if force or self._should_refresh():
            self._set_job_params(self._param_to_job_params())
            self.validate_parameters()

    def _should_refresh(self) -> bool:
        old_refresh_hash = self._refresh_hash
        new_refresh_hash = self._compute_refresh_hash()
        if old_refresh_hash != new_refresh_hash:
            self._refresh_hash = new_refresh_hash
            return True
        return False

    def _compute_refresh_hash(self) -> str:
        return sha256_hexdigest(self.to_dict(partial=True, validate=False))

    # ------------------------------------------------
    #                   Schema hooks
    # ------------------------------------------------

    @classmethod
    @mm.pre_load
    def _sanitize_param_pairs(cls, data: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        in_use_key = "param_pair_overrides"
        deprecated_keys = [
            "param_pairs",
            "param_set_pairs",
            "param_set_pair_overrides",
            "input_output_pairs",
            "input_output_pair_overrides",
        ]

        param_pairs = []

        if in_use_key in data:
            param_pairs.extend(data.pop(in_use_key))

        for k in deprecated_keys:
            if k in data:
                param_pairs.extend(data.pop(k))

        if param_pairs:
            data[in_use_key] = param_pairs

        return data

    @classmethod
    @mm.post_dump
    def _sanitize_params(cls, data: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        params = data["params"]
        for k, v in params.items():
            if isinstance(v, Resolvable):
                params[k] = v.to_str()
        return data



from dataclasses import dataclass
from typing import Any, List, cast

from aibs_informatics_core.models.base import SchemaModel, StringField, custom_field
from aibs_informatics_core.models.demand_execution.metadata import DemandExecutionMetadata
from aibs_informatics_core.models.demand_execution.parameters import DemandExecutionParameters
from aibs_informatics_core.models.demand_execution.platform import ExecutionPlatform
from aibs_informatics_core.models.demand_execution.resource_requirements import (
    DemandResourceRequirements,
)
from aibs_informatics_core.utils.hashing import sha256_hexdigest
from aibs_informatics_core.utils.json import JSON
from aibs_informatics_core.utils.time import get_current_time


@dataclass
class DemandExecution(SchemaModel):
    execution_type: str = custom_field()
    execution_id: str = custom_field()
    execution_image: str = custom_field(mm_field=StringField())
    execution_parameters: DemandExecutionParameters = custom_field(
        mm_field=DemandExecutionParameters.as_mm_field(),
        default_factory=DemandExecutionParameters,
    )
    execution_metadata: DemandExecutionMetadata = custom_field(
        mm_field=DemandExecutionMetadata.as_mm_field(),
        default_factory=DemandExecutionMetadata,
    )
    execution_platform: ExecutionPlatform = custom_field(
        mm_field=ExecutionPlatform.as_mm_field(), default_factory=ExecutionPlatform
    )
    resource_requirements: DemandResourceRequirements = custom_field(
        mm_field=DemandResourceRequirements.as_mm_field(),
        default_factory=DemandResourceRequirements,
    )

    def get_execution_hash(self, strict: bool = True) -> str:
        hash_components: List[Any] = [
            self.execution_type,
            self.execution_image,
            self.execution_parameters.command,
        ]
        if strict:
            hash_components.extend(
                [
                    self.execution_id,
                    self.execution_parameters.params,
                    self.execution_parameters.inputs,
                    self.execution_parameters.outputs,
                ]
            )
        return sha256_hexdigest(cast(JSON, hash_components))

    def generate_execution_name(self) -> str:
        """Creates a execution name for state machine execution

        Pattern (SUBJECT TO CHANGE): "{DEMAND_ID}-xxxxxx"

        The execution name is prefixed by demand Id to be more easily discoverable

        """
        return f'{self.execution_id}-{get_current_time().strftime("%Y%m%dT%H%M%S")}'



__all__ = [
    "BooleanField",
    "CustomAwareDateTime",
    "CustomStringField",
    "DictField",
    "EnumField",
    "FloatField",
    "FrozenSetField",
    "IntegerField",
    "ListField",
    "MappingField",
    "NestedField",
    "PathField",
    "RawField",
    "StringField",
    "TupleField",
    "UnionField",
    "UUIDField",
    "FieldMetadataBuilder",
    "FieldProps",
    "custom_field",
    "field_metadata",
    "ModelProtocol",
    "ModelBase",
    "BaseSchema",
    "DataClassModel",
    "MISSING",
    "NONE",
    "SchemaModel",
    "WithValidation",
    "post_dump",
    "pre_dump",
    "pre_load",
    "validates_schema",
]


from aibs_informatics_core.models.base.custom_fields import (
    BooleanField,
    CustomAwareDateTime,
    CustomStringField,
    DictField,
    EnumField,
    FloatField,
    FrozenSetField,
    IntegerField,
    ListField,
    MappingField,
    NestedField,
    PathField,
    RawField,
    StringField,
    TupleField,
    UnionField,
    UUIDField,
)
from aibs_informatics_core.models.base.field_utils import (
    FieldMetadataBuilder,
    FieldProps,
    custom_field,
    field_metadata,
)
from aibs_informatics_core.models.base.model import (
    MISSING,
    NONE,
    BaseSchema,
    DataClassModel,
    ModelBase,
    ModelProtocol,
    SchemaModel,
    WithValidation,
    post_dump,
    pre_dump,
    pre_load,
    validates_schema,
)



import re
from typing import ClassVar, Pattern

from aibs_informatics_core.collections import ValidatedStr


class SNSTopicArn(ValidatedStr):
    """ARN for AWS SNS Topic

    Examples:
        arn:aws:sns:us-west-2:123456789012:MyTopic
            region: us-west-2
            account: 123456789012
            topic: MyTopic

    """

    regex_pattern: ClassVar[Pattern] = re.compile(
        r"arn:aws:sns:([\w-]*):([\d]{12}):[\w\-\_]+(?:\.fifo)?"
    )



__all__ = [
    "AttributeBaseExpression",
    "ConditionBaseExpression",
    "ConditionBaseExpressionString",
    "ConditionBaseExpressionOrString",
]

import re
from dataclasses import dataclass
from typing import Any, ClassVar, List, Optional, Pattern, Union, cast

from aibs_informatics_core.collections import ValidatedStr
from aibs_informatics_core.models.base import (
    CustomStringField,
    ListField,
    RawField,
    SchemaModel,
    UnionField,
    custom_field,
)


@dataclass
class AttributeBaseExpression(SchemaModel):
    attr_class: str
    attr_name: Any = custom_field(mm_field=RawField())


@dataclass
class ConditionBaseExpression(SchemaModel):
    format: str = custom_field(mm_field=CustomStringField(str, strict_mode=True))
    operator: str = custom_field(mm_field=CustomStringField(str, strict_mode=True))
    # type: ignore[misc] # https://github.com/python/mypy/issues/731
    values: List[Union["ConditionBaseExpression", AttributeBaseExpression, Any]] = custom_field(
        mm_field=ListField(
            UnionField(
                [
                    (
                        lambda: ConditionBaseExpression,
                        lambda: ConditionBaseExpression.as_mm_field(),
                    ),
                    (AttributeBaseExpression, AttributeBaseExpression.as_mm_field()),
                    ((str, bool, int), RawField()),
                    ((list, set), ListField(RawField())),
                ]
            )
        )
    )


class ConditionBaseExpressionString(ValidatedStr):
    regex_pattern: ClassVar[Pattern] = re.compile(
        rf"([\w\.]+)(?:( begins_with | contains | NOT | IN |=|<>|<|<=|>|>=)(.+)|( attribute_exists))"
    )

    @property
    def condition_name(self) -> str:
        return self.get_match_groups()[0]

    @property
    def condition_operator(self) -> str:
        return cast(str, self.get_match_groups()[1] or self.get_match_groups()[3]).strip()

    @property
    def condition_format(self) -> str:
        operator = self.condition_operator
        # These mapping can be found here: boto3/dynamodb/conditions.py
        if operator in ["begins_with", "contains"]:
            return "{operator}({0}, {1})"
        elif operator == "attribute_exists":
            return "{operator}({0})"
        else:
            # operator is comparison condition
            return "{0} {operator} {1}"

    @property
    def condition_values(
        self,
    ) -> List[Union[ConditionBaseExpression, AttributeBaseExpression, Any]]:
        condition_values = [AttributeBaseExpression("Attr", self.condition_name)]
        value = self.get_match_groups()[2]

        def resolve(value: Optional[str], is_iterable: bool = False) -> Any:
            is_enclosed = lambda s, b, e: True if len(s) > 1 and (s[0], s[-1]) == (b, e) else False
            if not value:
                if is_iterable:
                    raise ValueError(
                        "IN operator must be followed by comma seperated values "
                        "enclosed in parentheses, e.g.'(optiona, optionb)'"
                    )
                return value
            if is_enclosed(value, "`", "`"):
                return value[1:-1]
            if is_iterable:
                if is_enclosed(value, "(", ")") or is_enclosed(value, "[", "]"):
                    value = value[1:-1]
                return [resolve(_.strip()) for _ in value.split(",")]
            elif value.lower() in ["false", "true"]:
                return value.lower() == "true"
            try:
                return int(value)
            except ValueError:
                pass
            try:
                return float(value)
            except ValueError:
                pass
            return value

        value = resolve(value, self.condition_operator == "IN")
        if value is not None:
            condition_values.append(value)
        return cast(
            List[Union[ConditionBaseExpression, AttributeBaseExpression, Any]], condition_values
        )

    def get_condition_expression(self, is_key: bool) -> ConditionBaseExpression:
        condition_format = self.condition_format
        condition_operator = self.condition_operator
        condition_values = self.condition_values
        if is_key:
            for condition_value in condition_values:
                if isinstance(condition_value, AttributeBaseExpression):
                    condition_value.attr_class = "Key"
        return ConditionBaseExpression(
            format=condition_format, operator=condition_operator, values=condition_values
        )


ConditionBaseExpressionOrString = Union[ConditionBaseExpression, ConditionBaseExpressionString]



import re
from typing import ClassVar, Optional
from urllib.parse import parse_qs

from aibs_informatics_core.collections import ValidatedStr
from aibs_informatics_core.models.aws.core import (
    AWS_ACCOUNT_PATTERN_STR,
    AWS_REGION_PATTERN_STR,
    AWSAccountId,
    AWSRegion,
)

LAMBDA_FUNCTION_NAME_PATTERN_STR = (
    r"(?:(?:(?:(?:arn:(?:aws[a-zA-Z-]*)?:lambda:)?"
    rf"(?:({AWS_REGION_PATTERN_STR}):)?)?"
    rf"(?:({AWS_ACCOUNT_PATTERN_STR}):)?)?"
    r"(?:function:))?"
    r"([a-zA-Z0-9\-_]{1,64})"
    r"(?::(\$LATEST|[a-zA-Z0-9-_]+))?"
)
LAMBDA_FUNCTION_URL_PATTERN_STR = (
    r"https:\/\/([a-zA-Z0-9-]+)\.lambda-url\."
    rf"({AWS_REGION_PATTERN_STR})"
    r"\.on\.aws"
    r"((?:\/[a-zA-Z0-9\._-]*)*)?"
    r"(?:\?(.*))?"
)

LAMBDA_FUNCTION_NAME_PATTERN = re.compile(LAMBDA_FUNCTION_NAME_PATTERN_STR)
LAMBDA_FUNCTION_URL_PATTERN = re.compile(LAMBDA_FUNCTION_URL_PATTERN_STR)


class LambdaFunctionName(ValidatedStr):
    regex_pattern: ClassVar[re.Pattern] = LAMBDA_FUNCTION_NAME_PATTERN

    @property
    def region(self) -> Optional[AWSRegion]:
        aws_region = self.get_match_groups()[0]
        return AWSRegion(aws_region) if aws_region else None

    @property
    def account_id(self) -> Optional[AWSAccountId]:
        aws_account_id = self.get_match_groups()[1]
        return AWSAccountId(aws_account_id) if aws_account_id else None

    @property
    def function_name(self) -> str:
        return self.get_match_groups()[2]

    @property
    def version(self) -> Optional[str]:
        return self.get_match_groups()[3]


class LambdaFunctionUrl(ValidatedStr):
    regex_pattern: ClassVar[re.Pattern] = LAMBDA_FUNCTION_URL_PATTERN

    @property
    def url_id(self) -> str:
        return self.get_match_groups()[0]

    @property
    def region(self) -> AWSRegion:
        aws_region = self.get_match_groups()[1]
        return AWSRegion(aws_region)

    @property
    def raw_path(self) -> Optional[str]:
        return self.get_match_groups()[-2] or None

    @property
    def raw_query(self) -> Optional[str]:
        return self.get_match_groups()[-1]

    @property
    def path(self) -> str:
        return self.raw_path or ""

    @property
    def query(self) -> dict[str, list[str]]:
        return parse_qs(self.raw_query) if self.raw_query else {}

    @property
    def base_url(self) -> str:
        return f"https://{self.url_id}.lambda-url.{self.region}.on.aws"



import re
from typing import ClassVar, Pattern

from aibs_informatics_core.collections import ValidatedStr


class EmailAddress(ValidatedStr):
    """Simple Email Address Validation

    No whitespaces, has an @
    """

    regex_pattern: ClassVar[Pattern] = re.compile(r"\S+@\S+")



__all__ = [
    "BEGINNING_OF_TIME",
    "get_current_time",
    "get_duration_in_secs",
    "from_isoformat_8601",
]

import datetime as dt
from typing import Optional

BEGINNING_OF_TIME = dt.datetime.fromtimestamp(0.0, tz=dt.timezone.utc)


def get_current_time(tz: Optional[dt.timezone] = None) -> dt.datetime:
    """Returns current time (in UTC) as a timezone aware datetime instance"""
    return dt.datetime.now(tz=tz or dt.timezone.utc)


def get_duration_in_secs(start: dt.datetime, stop: Optional[dt.datetime] = None) -> int:
    return round(((stop or get_current_time()) - start).total_seconds())


def from_isoformat_8601(iso8601_str: str) -> dt.datetime:
    """Convert ISO Format datetime string into a datetime object

    Example Strings:
        - 2022-06-09T06:58:14
        - 2022-06-09T06:58:14Z
        - 2022-06-09T06:58:14+11:00
        - 2022-06-09T06:58:14.000
        - 2022-06-09T06:58:14.000Z
        - 2022-06-09T06:58:14.000+11:00

    Args:
        iso8601_str (str): datetime string

    Returns:
        dt.datetime
    """
    fmt = "%Y-%m-%dT%H:%M:%S.%f" if "." in iso8601_str else "%Y-%m-%dT%H:%M:%S"

    if len(iso8601_str) < 6:
        raise ValueError(f"{iso8601_str} does not match {fmt}")
    elif iso8601_str[-6] in frozenset(("+", "-")):
        fmt = fmt + "%z"
    elif iso8601_str[-1] == "Z":
        fmt = fmt + "Z"
    return dt.datetime.strptime(iso8601_str, fmt)



from dataclasses import dataclass
from typing import Dict, List, Optional, Union

from aibs_informatics_core.models.aws.sfn import ExecutionArn
from aibs_informatics_core.models.base import SchemaModel, custom_field
from aibs_informatics_core.models.base.custom_fields import (
    BooleanField,
    CustomStringField,
    DictField,
    EnumField,
    ListField,
    StringField,
    UnionField,
)
from aibs_informatics_core.models.status import Status


@dataclass
class DemandExecutionMetadata(SchemaModel):
    user: Optional[str] = custom_field(default=None)
    arn: Optional[ExecutionArn] = custom_field(
        mm_field=CustomStringField(ExecutionArn), default=None
    )
    tag: Optional[str] = custom_field(default=None)
    notify_on: Optional[Dict[Status, bool]] = custom_field(
        mm_field=DictField(keys=EnumField(Status), values=BooleanField(), allow_none=True),
        default=None,
    )
    notify_list: Optional[List[str]] = custom_field(
        mm_field=ListField(StringField(), allow_none=True),
        default=None,
    )



import re
from functools import cached_property
from pathlib import Path
from typing import ClassVar, Optional, Union

from aibs_informatics_core.collections import ValidatedStr
from aibs_informatics_core.models.aws.core import AWS_ACCOUNT_PATTERN_STR as AWS_ACCOUNT_PATTERN
from aibs_informatics_core.models.aws.core import AWS_REGION_PATTERN_STR as AWS_REGION_PATTERN
from aibs_informatics_core.models.aws.core import AWSRegion
from aibs_informatics_core.models.base import CustomStringField

FILE_SYSTEM_ID_PATTERN = r"fs-[0-9a-f]{8,40}"
ACCESS_POINT_ID_PATTERN = r"fsap-[0-9a-f]{8,40}"


class FileSystemId(ValidatedStr):
    # https://docs.aws.amazon.com/efs/latest/ug/API_CreateAccessPoint.html#efs-CreateAccessPoint-request-FileSystemId
    regex_pattern: ClassVar[re.Pattern] = re.compile(
        rf"(arn:(aws[-a-z]*):elasticfilesystem:({AWS_REGION_PATTERN})(?::{AWS_ACCOUNT_PATTERN}):file-system/)?({FILE_SYSTEM_ID_PATTERN})"
    )

    @property
    def normalized(self) -> "FileSystemId":
        return FileSystemId(self.resource_id)

    @property
    def resource_id(self) -> str:
        return self.get_match_groups()[-1]

    @property
    def region(self) -> Optional[str]:
        return self.get_match_groups()[-2]

    @property
    def arn_prefix(self) -> Optional[str]:
        return self.get_match_groups()[0]


class AccessPointId(ValidatedStr):
    # https://docs.aws.amazon.com/efs/latest/ug/API_CreateAccessPoint.html#efs-CreateAccessPoint-response-AccessPointId
    regex_pattern: ClassVar[re.Pattern] = re.compile(
        rf"(arn:(aws[-a-z]*):elasticfilesystem:({AWS_REGION_PATTERN})(?::{AWS_ACCOUNT_PATTERN}):access-point/)?({ACCESS_POINT_ID_PATTERN})"
    )

    @property
    def normalized(self) -> str:
        return self.resource_id

    @property
    def resource_id(self) -> str:
        return self.get_match_groups()[-1]

    @property
    def region(self) -> Optional[str]:
        return self.get_match_groups()[-2]

    @property
    def arn_prefix(self) -> Optional[str]:
        return self.get_match_groups()[0]


class FileSystemDNSName(ValidatedStr):
    # https://docs.aws.amazon.com/efs/latest/ug/mounting-fs.html
    regex_pattern: ClassVar[re.Pattern] = re.compile(
        rf"({FILE_SYSTEM_ID_PATTERN})\.efs\.({AWS_REGION_PATTERN})\.amazonaws\.com"
    )

    @property
    def file_system_id(self) -> FileSystemId:
        return FileSystemId(self.get_match_groups()[0])

    @property
    def region(self) -> AWSRegion:
        return AWSRegion(self.get_match_groups()[1])

    @classmethod
    def build(cls, file_system_id: str, region: str) -> "FileSystemDNSName":
        return cls(f"{file_system_id}.efs.{region}.amazonaws.com")


PATH_PATTERN = r".*"


class EFSPath(ValidatedStr):
    """
    Custom EFS File URI

    EFS Does not have a standard URI format. This class provides a custom URI format for EFS
    resources. The format is as follows:

    [efs://](<fs-dns-name>|<resource-id>)[:]/<path>

        Where:
        - <fs-dns-name> = <fs-id>.efs.<region>.amazonaws.com
        - <resource-id> = <fs-id> | <fsap-id>
        - <path> = path to file or directory
        - <region> = aws region

    Examples (using fs-12345678 as file system id):

        Using efs:// prefix...
        - efs://fs-12345678.efs.us-east-1.amazonaws.com:/my/file.txt    -> /my/file.txt
        - efs://fs-12345678.efs.us-east-1.amazonaws.com/my/file.txt     -> /my/file.txt
        - efs://fs-12345678:/my/file.txt                                -> /my/file.txt
        - efs://fs-12345678/my/file.txt                                 -> /my/file.txt
        - efs://fs-12345678                                             -> /
        - efs://fs-12345678:                                            -> /
        - efs://fs-12345678/                                            -> /
        - efs://fs-12345678:/                                           -> /

        Using no prefix...
        - fs-12345678.efs.us-east-1.amazonaws.com:/my/file.txt          -> /my/file.txt
        - fs-12345678:/my/file.txt                                      -> /my/file.txt
        - fs-12345678:/                                                 -> /
        - fs-12345678:                                                  -> /


    """

    regex_pattern: ClassVar[re.Pattern] = re.compile(
        # rf"""
        # (?:efs:\/\/(?!.*:\/)|(?:(?:efs:\/\/)?(?=.*:\/)))
        # (?:({FILE_SYSTEM_ID_PATTERN})\.efs\.{AWS_REGION_PATTERN}\.amazonaws\.com|({FILE_SYSTEM_ID_PATTERN}))
        # :?
        # ({PATH_PATTERN})
        # """
        rf"""
        (?:
            (?:efs://)(?:({FILE_SYSTEM_ID_PATTERN})\.efs\.{AWS_REGION_PATTERN}\.amazonaws\.com|({FILE_SYSTEM_ID_PATTERN})):?
        |
            (?:efs://)?(?:({FILE_SYSTEM_ID_PATTERN})\.efs\.{AWS_REGION_PATTERN}\.amazonaws\.com|({FILE_SYSTEM_ID_PATTERN})):
        )
        ({PATH_PATTERN})
        """,
        re.VERBOSE,
    )

    @property
    def root(self) -> Path:
        return Path("/")

    @property
    def raw_path(self) -> str:
        return self.get_match_groups()[-1]

    @property
    def path(self) -> Path:
        return self.root / self.raw_path

    @cached_property
    def file_system_id(self) -> FileSystemId:
        groups = self.get_match_groups()
        file_store_id_str = groups[0] or groups[1] or groups[2] or groups[3]
        return FileSystemId(file_store_id_str)

    def as_uri(self) -> str:
        return f"efs://{self.file_system_id}:{self.path}"

    def as_dns_uri(self, region: AWSRegion) -> str:
        return f"efs://{FileSystemDNSName.build(self.file_system_id, region)}:{self.path}"

    @classmethod
    def build(
        cls, resource_id: Union[FileSystemId, FileSystemDNSName, str], path: Union[Path, str]
    ) -> "EFSPath":
        file_system_id: FileSystemId
        if FileSystemId.is_valid(resource_id):
            file_system_id = FileSystemId(resource_id).normalized
        elif FileSystemDNSName.is_valid(resource_id):
            file_system_id = FileSystemDNSName(resource_id).file_system_id
        else:
            raise ValueError(
                f"Invalid resource_id: {resource_id}. does not match file system id or dns name"
            )
        path = Path("/") / path
        return cls(f"{file_system_id}:{path.as_posix()}")

    @classmethod
    def as_mm_field(cls) -> CustomStringField:
        return CustomStringField(cls)

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, str) and EFSPath.is_valid(__value):
            return self.as_uri() == EFSPath(__value).as_uri()
        return super().__eq__(__value)

    def __truediv__(self, __other: Union[str, Path, "EFSPath"]) -> "EFSPath":
        """Appends a string or EFSPath key to the end of this EFSPath using the `/` operator

        Examples:
            >>> efs_uri = EFSPath("efs://fs-11111111/my-key") / "my-other-key"
            >>> assert efs_uri == "efs://fs-11111111/my-key/my-other-key"

            >>> another_efs_uri = EFSPath("efs://fs-11111111/key1") / EFSPath("efs://fs-22222222/key2")
            >>> assert another_efs_uri == "efs://fs-11111111/key1/key2"

        Args:
            __other (Union[str, EFSPath]): The key to append to the end of this EFSPath

        Returns:
            EFSPath: a new EFSPath with the appended key using the `/` operator
        """
        if isinstance(__other, EFSPath):
            __other = __other.path
        elif isinstance(__other, str):
            __other = Path(__other)

        if __other.is_relative_to("/"):
            __other = __other.relative_to("/")
        return EFSPath.build(resource_id=self.file_system_id, path=self.path / __other)

    def __rtruediv__(self, __other: str) -> "EFSPath":
        """Creates a new EFSPath by constructing a str or EFSPath key with this key

        Examples:
            >>> efs_uri = EFSPath("efs://fs-12345678/my-key") / "my-other-key"
            >>> assert efs_uri == "efs://fs-11111111/my-other-key"

            >>> another_efs_uri = EFSPath("efs://fs-11111111/key1") / EFSPath("fs-22222222/key2")
            >>> assert another_efs_uri == "efs://fs-11111111/key2"

        Args:
            __other (Union[str, EFSPath]): The key to append to the end of this EFSPath

        Returns:
            EFSPath: a new EFSPath with a new key
        """
        if EFSPath.is_valid(__other):
            return EFSPath(__other) / self
        elif FileSystemId.is_valid(__other):
            return EFSPath.build(FileSystemId(__other).normalized, path=self.path)
        elif FileSystemDNSName.is_valid(__other):
            return EFSPath.build(FileSystemDNSName(__other).file_system_id, path=self.path)
        else:
            return EFSPath.build(
                self.file_system_id, path=Path(__other) / self.path.relative_to("/")
            )

    def __floordiv__(self, __other: Union[str, Path, "EFSPath"]) -> "EFSPath":
        """Creates a new EFSPath by constructing a str or EFSPath file path with this fs id

        Examples:
            >>> efs_uri = EFSPath("efs://fs-11111111/my-key") // "my-other-key"
            >>> assert efs_uri == "efs://fs-11111111/my-other-key"

            >>> another_efs_uri = EFSPath("efs://fs-11111111/key1") // EFSPath("efs://fs-22222222/key2")
            >>> assert another_efs_uri == "efs://fs-11111111/key2"

        Args:
            __other (Union[str, EFSPath]): The key to append to the end of this EFSPath

        Returns:
            EFSPath: a new EFSPath with a new key
        """
        if isinstance(__other, EFSPath):
            __other = __other.path
        return EFSPath.build(resource_id=self.file_system_id, path=__other)



__all__ = [
    "remove_null_values",
    "remove_matching_values",
    "flatten_dict",
    "nested_dict",
    "convert_key_case",
]


from copy import deepcopy
from typing import Any, Callable, Dict, List, MutableMapping, Tuple, TypeVar, cast

T = TypeVar("T")
KT = TypeVar("KT")
VT = TypeVar("VT")


def remove_null_values(
    orig_dict: MutableMapping[KT, VT],
    in_place: bool = False,
    recursive: bool = False,
) -> MutableMapping[KT, VT]:
    """Removes null values from a dictionary object

    Args:
        orig_dict (MutableMapping[KT, VT]): target dictionary to prune null values from
        in_place (bool, optional): whether to do operation on target in place. Defaults to False.
        recursive (bool, optional): whether to prune recursively. Defaults to False.

    Returns:
        MutableMapping[KT, VT]: pruned dictionary
    """
    return remove_matching_values(
        orig_dict=orig_dict, in_place=in_place, recursive=recursive, target_value=None
    )


def remove_matching_values(
    orig_dict: MutableMapping[KT, VT],
    in_place: bool = False,
    recursive: bool = False,
    target_value: Any = None,
) -> MutableMapping[KT, VT]:
    """Filter out keys with values matching target value

    Args:
        orig_dict (MutableMapping[KT, VT]): Original dictionary
        in_place (bool, optional): If true, removes keys in place. Defaults to False.
        recursive (bool, optional): If true, recursively removes values if target value.
            Defaults to False.

    Returns:
        MutableMapping[KT, VT]: dictionary without target values
    """
    filtered_dict = orig_dict if in_place else deepcopy(orig_dict)

    for k, v in list(filtered_dict.items()):
        if isinstance(v, dict) and recursive:
            filtered_dict[k] = cast(VT, remove_null_values(v, in_place=True, recursive=recursive))
        elif v == target_value:
            filtered_dict.pop(k)
    return filtered_dict


def flatten_dict(
    data: MutableMapping[str, Any], parent_key: str = "", delimiter: str = "."
) -> Dict[str, Any]:
    """
    Flattens a nested dictionary by concatenating the keys with a given delimiter.

    Args:
    data (Dict[str, Any]): The dictionary to flatten.
    parent_key (str, optional): The concatenated key for the nested items. Defaults to ''.
    delimiter (str, optional): The delimiter to use for concatenating the keys. Defaults to '.'.

    Returns:
    Dict[str, Any]: The flattened dictionary.
    """
    items: List[Tuple[str, Any]] = []
    for key, value in data.items():
        new_key = f"{parent_key}{delimiter}{key}" if parent_key else key
        if isinstance(value, dict):
            items.extend(flatten_dict(value, new_key, delimiter).items())
        else:
            items.append((new_key, value))
    return dict(items)


def nested_dict(data: MutableMapping[str, Any], delimiter: str = ".") -> Dict[str, Any]:
    """
    Creates a nested dictionary by splitting the keys by a given delimiter.

    Args:
    data (Dict[str, Any]): The dictionary to transform.
    delimiter (str, optional): The delimiter to use for splitting the keys. Defaults to '.'.

    Returns:
    Dict[str, Any]: The transformed, nested dictionary.

    Raises:
    ValueError: If a collision is detected during the transformation process.
    """
    result: Dict[str, Any] = {}
    for key, value in data.items():
        parts = key.split(delimiter)
        d = result
        for part in parts[:-1]:
            if part not in d:
                d[part] = {}
            elif not isinstance(d[part], dict):
                raise ValueError(f"Key collision detected at '{part}' when processing '{key}'")
            d = d[part]
        if parts[-1] in d and d[parts[-1]] != value:
            raise ValueError(f"Key collision detected at '{parts[-1]}' when processing '{key}'")
        d[parts[-1]] = value
    return result


def convert_key_case(data: T, key_case: Callable[[str], str]) -> T:
    if isinstance(data, dict):
        return {
            (key_case(k) if isinstance(k, str) else k): convert_key_case(v, key_case)
            for k, v in data.items()
        }  # type: ignore
    elif isinstance(data, list):
        return [convert_key_case(item, key_case) for item in data]  # type: ignore
    else:
        return data



__all__ = [
    "ModelProtocol",
    "ModelBase",
    "BaseSchema",
    "DataClassModel",
    "MISSING",
    "NONE",
    "SchemaModel",
    "WithValidation",
    "post_dump",
    "pre_dump",
    "pre_load",
    "validates_schema",
]

import abc
import inspect
import json
from dataclasses import MISSING as dataclass_MISSING
from dataclasses import Field, dataclass, fields
from functools import wraps
from pathlib import Path
from types import MethodType
from typing import (
    Any,
    ClassVar,
    Dict,
    Optional,
    Protocol,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    get_type_hints,
    runtime_checkable,
)

import marshmallow as mm
import yaml  # type: ignore[import-untyped]
from dataclasses_json import DataClassJsonMixin, Undefined, config
from dataclasses_json.core import _ExtendedEncoder
from marshmallow import post_dump, pre_dump, pre_load, validates_schema
from marshmallow.decorators import POST_LOAD

from aibs_informatics_core.models.base.custom_fields import NestedField
from aibs_informatics_core.models.base.field_utils import FieldProps
from aibs_informatics_core.utils.decorators import cache
from aibs_informatics_core.utils.json import JSONObject
from aibs_informatics_core.utils.tools.dicttools import remove_matching_values

T = TypeVar("T")


DEFAULT_PARTIAL = False
DEFAULT_VALIDATE = True

M = TypeVar("M", bound="ModelBase")
DCM = TypeVar("DCM", bound="DataClassModel")
SM = TypeVar("SM", bound="SchemaModel")


# --------------------------------------------------------------
#                             ModelProtocol
# --------------------------------------------------------------


@runtime_checkable
class ModelProtocol(Protocol):
    @classmethod
    def from_dict(cls: Type[T], data: JSONObject, **kwargs) -> T:
        ...  # pragma: no cover

    def to_dict(self, **kwargs) -> JSONObject:
        ...  # pragma: no cover

    @classmethod
    def from_json(cls: Type[T], data: str, **kwargs) -> T:
        ...  # pragma: no cover

    def to_json(self, **kwargs) -> str:
        ...  # pragma: no cover

    @classmethod
    def from_path(cls: Type[T], path: Path, **kwargs) -> T:
        ...  # pragma: no cover

    def to_path(self, path: Path, **kwargs):
        ...  # pragma: no cover


# --------------------------------------------------------------
#                             BaseModel ABC
# --------------------------------------------------------------


class ModelBase:
    @classmethod
    @abc.abstractmethod
    def from_dict(cls: Type[M], data: JSONObject, **kwargs) -> M:
        raise NotImplementedError(
            f"Must implement this method in {cls.__name__}"
        )  # pragma: no cover

    @abc.abstractmethod
    def to_dict(self, **kwargs) -> JSONObject:
        raise NotImplementedError(
            f"Must implement this method in {self.__class__.__name__}"
        )  # pragma: no cover

    @classmethod
    def from_json(cls: Type[M], data: str, **kwargs) -> M:
        return cls.from_dict(json.loads(data), **kwargs)

    def to_json(self, **kwargs) -> str:
        return json.dumps(self.to_dict(**kwargs), indent=4)

    @classmethod
    def from_path(cls: Type[M], path: Path, **kwargs) -> M:
        if path.suffix in (".yml", ".yaml"):
            path.read_text()
            with open(path, "r") as f:
                return cls.from_dict(yaml.safe_load(f), **kwargs)
        else:
            return cls.from_dict(json.loads(path.read_text()), **kwargs)

    def to_path(self, path: Path, **kwargs):
        path.write_text(self.to_json(**kwargs))


# --------------------------------------------------------------
#                        DataClassModel
# --------------------------------------------------------------


class DataClassModel(DataClassJsonMixin, ModelBase):
    dataclass_json_config: ClassVar[dict] = config(  # type: ignore[misc]
        undefined=Undefined.EXCLUDE,  # default behavior for handling undefined fields
        exclude=lambda f: f is None,  # excludes values if None by default
    )["dataclasses_json"]

    @classmethod
    def from_dict(  # type: ignore[override]
        cls: Type[DCM], data: JSONObject, partial: bool = DEFAULT_PARTIAL, **kwargs
    ) -> DCM:
        return super().from_dict(data, infer_missing=partial)

    def to_dict(self, partial: bool = DEFAULT_PARTIAL, **kwargs) -> JSONObject:  # type: ignore[override]
        return super().to_dict(encode_json=kwargs.get("encode_json", True))

    @classmethod
    def from_json(cls: Type[DCM], data: str, **kwargs) -> DCM:  # type: ignore[override]
        return cls.from_dict(json.loads(data), **kwargs)

    def to_json(self, **kwargs) -> str:
        json_encoder_cls: Type[json.JSONEncoder] = kwargs.pop("json_encoder", _ExtendedEncoder)
        return json.dumps(self.to_dict(**kwargs), indent=4, cls=json_encoder_cls)

    @classmethod
    @cache
    def get_model_fields(cls) -> Tuple[Field, ...]:
        resolved: Optional[Dict[str, Type]] = None
        class_fields = fields(cls)  # type: ignore[arg-type]
        for f in class_fields:
            # There is a bug with https://peps.python.org/pep-0563/
            # (postponed evaluation of annotations) where the type of the field is a string.
            # This is a workaround for that issue.
            # https://stackoverflow.com/a/55938344/4544508
            if isinstance(f.type, str):
                # This avoids running `get_type_hints` unnecessarily
                if resolved is None:
                    resolved = get_type_hints(cls)
                f.type = resolved[f.name]

        return class_fields  # type: ignore[arg-type]

    @classmethod
    def is_missing(cls, value: Any) -> bool:
        return value is dataclass_MISSING or value is MISSING or value is ...


# --------------------------------------------------------------
#                      ValidatedModel
# --------------------------------------------------------------


MISSING = mm.missing


class WithValidation:
    """Provides Validations for model for (/de-)serialization"""

    @classmethod
    @abc.abstractmethod
    def validate(cls, data: Dict[str, Any], partial: bool = DEFAULT_PARTIAL, **kwargs):
        raise NotImplementedError("Must implement")  # pragma: no cover


# --------------------------------------------------------------
#                     SchemaModel
# --------------------------------------------------------------


class SchemaModel(DataClassModel):
    _schema_config: ClassVar[Dict[str, Any]] = {}

    def __init_subclass__(cls: Type[SM], **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        if cls._schema_config.get("attach_schema_hooks", True):
            attach_schema_hooks(cls, cls._schema_config.get("remove_post_load_hooks", True))

    @classmethod
    def from_dict(  # type: ignore[override]
        cls: Type[SM], data: Dict[str, Any], partial: bool = DEFAULT_PARTIAL, **kwargs
    ) -> SM:
        """deserialize JSON data (and validate)

        Args:
            cls (Type[SM]): class type to construct from data
            data (JSONObject): JSON data that will be deserialized
            partial (bool, optional): Whether to partially construct. Defaults to False.

        Returns:
            SM: _description_
        """
        # return super().from_dict(data, partial=partial, **kwargs)
        return cast(
            SM, cls.model_schema(partial=partial, **kwargs).load(data=data, partial=partial)
        )

    def to_dict(  # type: ignore[override]
        self, partial: bool = DEFAULT_PARTIAL, validate: bool = DEFAULT_VALIDATE, **kwargs
    ) -> JSONObject:
        model_dict = cast(JSONObject, self.model_schema(partial=partial).dump(self, **kwargs))
        if validate:
            self.validate(model_dict, partial=partial, **kwargs)
        return model_dict

    def update(self, data: Dict[str, Any]):
        """Update dataclass attributes with validation"""
        existing_data = self.to_dict(partial=True)
        existing_data.update(data)
        updated_self = self.from_dict(existing_data, partial=True)

        for field in self.get_model_fields():
            new_value = getattr(updated_self, field.name)

            if FieldProps(field).requires_init():
                if new_value is not None or new_value == mm.missing:
                    setattr(self, field.name, new_value)
            else:
                setattr(self, field.name, new_value)

    def validate_obj(self, partial: bool = DEFAULT_PARTIAL, **kwargs):
        self.validate(self, partial=partial, **kwargs)

    @classmethod
    def validate(
        cls: Type[SM], data: Union[Dict[str, Any], SM], partial: bool = DEFAULT_PARTIAL, **kwargs
    ):
        try:
            if isinstance(data, cls):
                data = data.to_dict(partial=partial)
            cls.from_dict(cast(JSONObject, data), partial=partial)
        except Exception as e:
            raise mm.ValidationError(f"Validation failed for {cls.__name__}: {e}")

    def is_valid_obj(self, partial: bool = DEFAULT_PARTIAL, **kwargs) -> bool:
        return self.is_valid(self, partial=partial, **kwargs)

    @classmethod
    def is_valid(
        cls: Type[SM], data: Union[Dict[str, Any], SM], partial: bool = DEFAULT_PARTIAL, **kwargs
    ) -> bool:
        try:
            cls.validate(data, partial=partial, **kwargs)
        except mm.ValidationError:
            return False
        return True

    @classmethod
    def empty(cls: Type[SM]) -> SM:
        empty: Dict[str, Any] = {}
        return cls.from_dict(empty, partial=True)

    # ----------------------------------------
    # Schema-related fields
    # ----------------------------------------

    @classmethod
    def as_mm_field(cls, partial: bool = DEFAULT_PARTIAL, **kwargs) -> NestedField:
        return NestedField(cls.model_schema(partial=partial, **kwargs))

    @classmethod
    def model_schema(cls, partial: bool = DEFAULT_PARTIAL, **kwargs) -> mm.Schema:
        """Marshmallow Schema representing this class.

        By default, a schema is generated for you.

        """
        return cls.schema(partial=partial, **kwargs)

    # ----------------------------------------
    # Default schema hooks
    # ----------------------------------------

    @classmethod
    @mm.post_load
    def make_object(
        cls: Type[T], data: Dict[str, Any], partial: bool = DEFAULT_PARTIAL, **kwargs
    ) -> T:
        """Method for defining how to make an instance based on validated data.

        Args:
            cls (Type[T]): The class of instance to make
            partial (bool, optional): Whether to partially construct. Defaults to False.
            data (dict): Validated data (based on schema)

        Returns:
            T: An instance of the model
        """
        if partial:
            class_fields = fields(cls)  # type: ignore[arg-type]
            for class_field in class_fields:
                if class_field.name not in data:
                    field_props = FieldProps(class_field)
                    if not field_props.requires_init():
                        continue
                    elif field_props.is_optional_type():
                        data[class_field.name] = None
                    else:
                        # required value. We will do something whacky
                        data[class_field.name] = MISSING

        return cls(**data)  # type: ignore

    @classmethod
    @mm.post_dump
    def remove_optional_values(cls, data: dict, **kwargs) -> dict:
        """Schema post-dump hook which removes optional null values from data

        Args:
            data (dict): self as dict
        """

        class_fields = cls.get_model_fields()
        for class_field in class_fields:
            if class_field.name in data:
                field_props = FieldProps(class_field)
                if field_props.is_optional_type() and data[class_field.name] is None:
                    data.pop(class_field.name)
        return data

    @classmethod
    @mm.post_dump
    def remove_missing_values(cls, data: dict, **kwargs) -> dict:
        """Schema Post dump hook that removes MISSING values from data"""
        new_data = remove_matching_values(data, recursive=True, target_value=MISSING)
        return cast(dict, new_data)

    def is_partial(self) -> bool:
        """Verifies if object is partially defined"""
        for class_field in self.get_model_fields():
            field_props = FieldProps(class_field)
            if not field_props.is_optional_type() and getattr(self, class_field.name) is MISSING:
                return True
        return False

    def copy(self: SM, partial: bool = DEFAULT_PARTIAL, **kwargs) -> SM:
        return self.from_dict(
            data=self.to_dict(partial=partial, **kwargs), partial=partial, **kwargs
        )


class ModelSchemaMethod(Protocol):
    def __call__(self, cls: Type[SM], partial: bool, **kwargs) -> mm.Schema:
        ...  # pragma: no cover


class ModelClassMethod(Protocol):
    def __call__(*args, **kwargs) -> Any:
        ...  # pragma: no cover


def attach_schema_hooks(cls: Type[SchemaModel], remove_post_load_hooks: bool = True):
    """Attaches schema hooks from SchemaModel class onto the schema class

    Args:
        cls (Type[SchemaModel]): The subclass of schema model
        remove_post_load_hooks (bool): Whether to remove post_load methods from schema.
            Defaults to True.
    """
    MARSHMALLOW_HOOK_ATTR = "__marshmallow_hook__"

    # We must use the underlying __func__ of this method variable in order to pass in
    # the correct class variable. This was problematic for classes that subclassing
    # child classes of SchemaModel.
    # In the following scenario:
    #   B -child-of--> A -child-of--> SchemaModel
    #
    #   B.model_schema() would call the bound method of A.model_schema
    #   because we were wrapping the bound method, not the function
    model_schema_method: ModelSchemaMethod = cls.model_schema.__func__  # type: ignore

    def check_and_attach_schema_hook(
        cls: Type[SM],
        schema: mm.Schema,
        class_method_name: str,
        class_method: ModelClassMethod,
    ):
        """Decorates a SchemaModel class method as a mm.Schema hook and attaches to the schema

        Args:
            cls (Type[SM]): subclass of schema model
            schema (mm.Schema): The schema instance of the schema model
            class_method_name (str): name of class method being decorated, attached
            class_method (ModelClassMethod): class method being decorated, attached

        """
        try:
            hook_config = class_method.__marshmallow_hook__  # type: ignore[attr-defined]
        except AttributeError:
            return

        @wraps(class_method)
        def schema_hook(self, *args, **kwargs):
            return class_method(*args, **kwargs)

        setattr(schema_hook, MARSHMALLOW_HOOK_ATTR, hook_config)
        setattr(schema.__class__, f"_{class_method_name}__auto", MethodType(schema_hook, schema))

    @cache
    @wraps(model_schema_method)
    def model_schema_with_hooks(
        cls: Type[SM], partial: bool = DEFAULT_PARTIAL, **kwargs
    ) -> mm.Schema:
        schema = model_schema_method(cls, partial=partial, **kwargs)

        if remove_post_load_hooks:
            post_load_key = POST_LOAD
            for post_load_method_name, many, args in schema._hooks.get(post_load_key, []):
                try:
                    post_load_method = getattr(schema, post_load_method_name)
                    if hook_attr := getattr(post_load_method, MARSHMALLOW_HOOK_ATTR):
                        # Ignore auto-generated post_load methods
                        if post_load_method_name == f"_{cls.make_object.__name__}__auto":
                            continue
                        hook_attr.pop(post_load_key, None)
                except:
                    pass

        class_methods = inspect.getmembers(cls, predicate=inspect.ismethod)

        for class_method_name, class_method in class_methods:
            check_and_attach_schema_hook(cls, schema, class_method_name, class_method)

        # Lastly, reset hooks of the mm.Schema by re-running `resolve_hooks`
        setattr(schema.__class__, "_hooks", schema.__class__.resolve_hooks())
        return schema

    setattr(cls, "model_schema", MethodType(model_schema_with_hooks, cls))


class BaseSchema(mm.Schema):
    pass


@dataclass
class NONE(SchemaModel):
    """Can be used as a No-op"""



# needed for python 3.11 to enable postponed evaluation of annotations
from __future__ import annotations

import aibs_informatics_core

__all__ = [
    "ApiRoute",
    "ApiClientInterface",
    "ApiRequestConfig",
    "ClientRouteMethod",
    "BoundClientRouteMethod",
]

import logging
import re
import sys
from abc import abstractmethod
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    List,
    Match,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

import requests  # type: ignore[import-untyped]
from requests.auth import AuthBase  # type: ignore[import-untyped]

from aibs_informatics_core.exceptions import ValidationError
from aibs_informatics_core.models.api.http_parameters import HTTPParameters
from aibs_informatics_core.models.base import (
    CustomStringField,
    ModelProtocol,
    SchemaModel,
    custom_field,
)
from aibs_informatics_core.models.version import VersionStr
from aibs_informatics_core.utils.json import JSON
from aibs_informatics_core.utils.os_operations import get_env_var
from aibs_informatics_core.utils.tools.strtools import removesuffix
from aibs_informatics_core.utils.version import get_version

if sys.version_info >= (3, 10):
    from inspect import get_annotations
else:  # pragma: no cover
    import types
    from typing import Callable, Mapping, Optional, Union

    # not supported in 3.9
    def get_annotations(
        obj: Union[Callable[..., object], Type[Any], types.ModuleType],
        *,
        globals: Optional[Mapping[str, Any]] = None,
        locals: Optional[Mapping[str, Any]] = None,
        eval_str: bool = False,
    ) -> Dict[str, Any]:
        return obj.__annotations__


T = TypeVar("T")
AuthType = TypeVar("AuthType", bound=AuthBase)

logger = logging.getLogger(__name__)


API_SERVICE_LOG_LEVEL_ENV_VAR = "API_LOG_LEVEL"
API_SERVICE_LOG_LEVEL_KEY = "x-log-level"

CLIENT_VERSION_ENV_VAR = "API_CLIENT_VERSION"
CLIENT_VERSION_KEY = "x-client-version"

CLIENT_VERSION_PACKAGE_ENV_VAR = "API_CLIENT_VERSION_PACKAGE"

DYNAMIC_ROUTE_PATTERN = re.compile(r"(<(\w+)>)")


API_REQUEST = TypeVar("API_REQUEST", bound=ModelProtocol)
API_RESPONSE = TypeVar("API_RESPONSE", bound=ModelProtocol)


@dataclass
class ApiRequestConfig(SchemaModel):
    client_version: VersionStr = custom_field(mm_field=CustomStringField(VersionStr))
    service_log_level: Optional[str] = custom_field(default=None)

    client_version_default: ClassVar[Optional[VersionStr]] = None
    client_version_package_name_default: ClassVar[str] = aibs_informatics_core.__name__

    def to_headers(self) -> Dict[str, str]:
        headers: Dict[str, str] = {
            CLIENT_VERSION_KEY: str(self.client_version),
        }
        if self.service_log_level:
            headers[API_SERVICE_LOG_LEVEL_KEY] = self.service_log_level

        return headers

    @classmethod
    def from_headers(cls, headers: Dict[str, str]) -> ApiRequestConfig:
        return cls.from_dict(
            dict(
                client_version=headers.get(CLIENT_VERSION_KEY),
                service_log_level=headers.get(API_SERVICE_LOG_LEVEL_KEY),
            )
        )

    @classmethod
    def build(cls, **kwargs) -> ApiRequestConfig:
        client_version = cls.build__client_version(**kwargs)
        service_log_level = get_env_var(API_SERVICE_LOG_LEVEL_ENV_VAR)

        return cls(
            client_version=client_version,
            service_log_level=service_log_level,
        )

    @classmethod
    def build__client_version(cls, **kwargs) -> VersionStr:
        """Builds the client version

        Order of precedence:
        1. "client_version" in kwargs
        2. "client_version_package_name" in kwargs -> get_version(VALUE)
        3. client_version_default (if set) in class
        4. client_version_package_name_default in class -> get_version(VALUE)

        Returns:
            VersionStr: version string resolved from the above order of precedence
        """
        if "client_version" in kwargs:
            return VersionStr(kwargs["client_version"])
        elif "client_version_package_name" in kwargs:
            return VersionStr(get_version(kwargs["client_version_package_name"]))
        elif cls.client_version_default:
            return cls.client_version_default
        elif cls.client_version_package_name_default:
            return VersionStr(get_version(cls.client_version_package_name_default))
        else:  # pragma: no cover
            raise ValueError("Could not resolve client version")


class ApiHeadersMixin:
    request_config_cls: ClassVar[Type[ApiRequestConfig]] = ApiRequestConfig

    @classmethod
    def generate_headers(cls) -> Dict[str, str]:
        """Custom headers attached with each client request

        Returns:
            Dict[str, Any]: dictionary of header key-value pairs
        """
        return cls.request_config_cls.build().to_headers()

    @classmethod
    def validate_headers(
        cls,
        headers: Dict[str, str],
        minimum_client_version: Optional[Union[VersionStr, str]] = None,
    ) -> None:
        config = cls.resolve_request_config(headers)

        # validate client version
        if minimum_client_version:
            minimum_client_version = VersionStr(minimum_client_version)
            if config.client_version < minimum_client_version:
                raise ValueError(
                    f"Client version {config.client_version} is too old. "
                    f"Must use client version {minimum_client_version} or newer"
                )

    @classmethod
    def resolve_request_config(cls, headers: Dict[str, str]) -> ApiRequestConfig:
        return cls.request_config_cls.from_headers(headers)


class ApiRoute(Generic[API_REQUEST, API_RESPONSE], ApiHeadersMixin):
    @classmethod
    @abstractmethod
    def route_rule(cls) -> str:
        """The rule for the given API route

        Returns:
            str: a rule (e.g. /prefix/route/<execution_id>)
        """
        raise NotImplementedError("Specify a rule for the given API route")  # pragma: no cover

    @classmethod
    @abstractmethod
    def route_method(cls) -> Union[str, List[str]]:
        """Specifies the methods available

        Returns:
            Union[str, List[str]]: e.g. ['POST', 'PUT'], 'GET'
        """
        raise NotImplementedError("Specify a method for the given API route")  # pragma: no cover

    @classmethod
    def primary_route_method(cls) -> str:
        method = cls.route_method()
        return next(iter(method)) if isinstance(method, list) else method

    @classmethod
    def route_rule_param_keys(cls) -> List[str]:
        return [key for (_, key) in DYNAMIC_ROUTE_PATTERN.findall(cls.route_rule())]

    @classmethod
    def service_name(cls) -> str:
        return cls.__name__

    @classmethod
    def get_request_cls(cls) -> Type[API_REQUEST]:
        return cls.__orig_bases__[0].__args__[0]  # type: ignore

    @classmethod
    def get_response_cls(cls) -> Type[API_RESPONSE]:
        return cls.__orig_bases__[0].__args__[1]  # type: ignore

    @classmethod
    def get_client_method_name(cls) -> str:
        client_method_name = re.sub(
            "((?<=[a-z])[A-Z]|(?!^)[A-Z](?=[a-z]))", r"_\1", cls.__name__
        ).lower()
        return removesuffix(client_method_name, "_route")

    @classmethod
    def get_http_request(
        cls, request: API_REQUEST, base_url: str, **request_kwargs
    ) -> requests.Request:
        """

        Args:
            request (API_REQUEST): _description_
            base_url (str): _description_

        Returns:
            requests.Request: _description_
        """

        http_parameters = cls.get_http_parameters_from_request(request)

        route_path = cls._get_route_path(http_parameters)
        url = f"{base_url}{route_path}"

        if "headers" not in request_kwargs:
            request_kwargs["headers"] = cls.generate_headers()

        http_request = requests.Request(
            method=cls.primary_route_method(),
            url=url,
            json=http_parameters.stringified_request_body,
            params=http_parameters.stringified_query_params,
            **request_kwargs,
        )
        return http_request

    @classmethod
    def get_request_from_http_parameters(cls, http_parameters: HTTPParameters) -> API_REQUEST:
        request_cls = cls.get_request_cls()
        return request_cls.from_dict(http_parameters.merged_params)

    @classmethod
    def get_http_parameters_from_request(cls, request: API_REQUEST) -> HTTPParameters:
        request_body = cast(Dict[str, JSON], request.to_dict())

        # Initialize the two parameter types
        route_params = {}
        query_params = {}

        for route_param_key in cls.route_rule_param_keys():
            if route_param_key not in request_body:
                raise ValidationError(
                    f"{route_param_key} specified in {cls.route_rule()} is not a valid field"
                    f"in the request class {request.__class__}"
                )
            route_params[route_param_key] = request_body.pop(route_param_key)

        method = cls.primary_route_method()
        if method == "GET":
            query_params = request_body.copy()
            request_body = {}
        return HTTPParameters(
            route_params=route_params,
            query_params=query_params,
            request_body=request_body,
        )

    @classmethod
    def _get_route_path(cls, http_parameters: HTTPParameters) -> str:
        str_route_parameters = http_parameters.stringified_route_params

        def _stringify_field(dynamic_route_match: Match) -> str:
            field_name = dynamic_route_match.group(2)
            try:
                return str_route_parameters[field_name]
            except Exception as e:
                raise ValidationError(
                    f"Couldn't resolve {field_name} from {str_route_parameters}"
                ) from e

        return DYNAMIC_ROUTE_PATTERN.sub(repl=_stringify_field, string=cls.route_rule())

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(rule={self.route_rule()}, method={self.route_method()})"

    @classmethod
    def create_route_method(cls) -> ClientRouteMethod[API_REQUEST, API_RESPONSE]:
        def route_method(self: ApiClientInterface, request: API_REQUEST) -> API_RESPONSE:
            return self.call(cls(), request)

        return cast(ClientRouteMethod[API_REQUEST, API_RESPONSE], route_method)


class ApiClientInterface(Protocol):
    def call(
        self, route: ApiRoute[API_REQUEST, API_RESPONSE], request: API_REQUEST
    ) -> API_RESPONSE:
        ...  # pragma: no cover


# Solution for adding class methods to client class with named arguments
# https://github.com/python/typing/discussions/1040
class BoundClientRouteMethod(Protocol, Generic[API_REQUEST, API_RESPONSE]):  # type: ignore
    def __call__(self, request: API_REQUEST) -> API_RESPONSE:
        ...  # pragma: no cover


class ClientRouteMethod(Protocol, Generic[API_REQUEST, API_RESPONSE]):  # type: ignore
    def __call__(__self, self: ApiClientInterface, request: API_REQUEST) -> API_RESPONSE:
        ...  # pragma: no cover

    @overload
    def __get__(
        self, obj: ApiClientInterface, objtype: Optional[Type[ApiClientInterface]] = None
    ) -> BoundClientRouteMethod[API_REQUEST, API_RESPONSE]:
        ...  # pragma: no cover

    @overload
    def __get__(  # type: ignore[misc]
        self, obj: None, objtype: Optional[Type[ApiClientInterface]] = None
    ) -> ClientRouteMethod[API_REQUEST, API_RESPONSE]:
        ...  # pragma: no cover

    def __get__(
        self,
        obj: Optional[ApiClientInterface],
        objtype: Optional[Type[ApiClientInterface]] = None,
    ) -> Union[
        BoundClientRouteMethod[API_REQUEST, API_RESPONSE],
        ClientRouteMethod[API_REQUEST, API_RESPONSE],
    ]:
        ...  # pragma: no cover



__all__ = [
    "ENV_BASE_KEY",
    "ENV_BASE_KEY_ALIAS",
    "ENV_LABEL_KEY",
    "ENV_LABEL_KEY_ALIAS",
    "ENV_TYPE_KEY",
    "ENV_TYPE_KEY_ALIAS",
    "LABEL_KEY",
    "LABEL_KEY_ALIAS",
    "EnvBase",
    "EnvBaseEnumMixins",
    "EnvBaseMixins",
    "EnvType",
    "ResourceNameBaseEnum",
    "SupportedDelim",
    "get_env_base",
    "get_env_type",
    "get_env_label",
]

import re
from enum import Enum
from typing import Generic, Literal, Optional, Tuple, Type, TypeVar, Union, overload

from aibs_informatics_core.collections import StrEnum, ValidatedStr
from aibs_informatics_core.exceptions import ApplicationException
from aibs_informatics_core.utils.os_operations import get_env_var, set_env_var

ENV_BASE_KEY = "env_base"
ENV_LABEL_KEY = "env_label"
ENV_TYPE_KEY = "env_type"
LABEL_KEY = "label"

ENV_BASE_KEY_ALIAS = ENV_BASE_KEY.upper()
ENV_LABEL_KEY_ALIAS = ENV_LABEL_KEY.upper()
ENV_TYPE_KEY_ALIAS = ENV_TYPE_KEY.upper()
LABEL_KEY_ALIAS = LABEL_KEY.upper()


class EnvType(StrEnum):
    DEV = "dev"
    TEST = "test"
    PROD = "prod"


SupportedDelim = Literal["-", "_", ":", "/"]
E = TypeVar("E", bound="EnvBase")


class EnvBase(ValidatedStr):
    ENV_BASE_KEY = ENV_BASE_KEY
    ENV_LABEL_KEY = ENV_LABEL_KEY
    ENV_TYPE_KEY = ENV_TYPE_KEY
    regex_pattern = re.compile(f"({'|'.join(EnvType.values())})" r"(?:-(\w+)|)")

    @property
    def env_type(self) -> EnvType:
        return self.to_type_and_label()[0]

    @property
    def env_label(self) -> Optional[str]:
        return self.to_type_and_label()[1]

    def get_construct_id(self, *names: str) -> str:
        return self.prefixed(*names)

    def get_stack_name(self, *names: str) -> str:
        return self.prefixed(*names)

    def get_stage_name(self, *names: str) -> str:
        return self.prefixed(*names)

    def get_pipeline_name(self, *names: str) -> str:
        return self.prefixed(*names)

    def get_repository_name(self, *names: str) -> str:
        return self.concat(self, self.concat(*names), delim="/")

    def get_resource_name(self, *names: str) -> str:
        return self.prefixed(*names)

    def get_table_name(self, name: str) -> str:
        return self.prefixed(name)

    def get_function_name(self, name: str) -> str:
        return self.prefixed(name)

    def get_state_machine_name(self, name: str) -> str:
        return self.get_resource_name(name)

    def get_job_name(self, name: str, hash_value: str) -> str:
        return self.get_resource_name(name, hash_value)

    def get_state_machine_log_group_name(self, name: str) -> str:
        return self.concat(
            # https://docs.aws.amazon.com/step-functions/latest/dg/bp-cwl.html
            "/aws/vendedlogs",
            self,
            "states",
            name,
            delim="/",
        )

    def get_metric_namespace(self, name: str) -> str:
        return self.concat("AIBS", self, name, delim="/")

    def get_bucket_name(self, base_name: str, account_id: str, region: str) -> str:
        return self.concat(self, base_name, region, account_id, delim="-")

    def get_ssm_param_name(self, *names: str) -> str:
        return self.concat("", self, self.concat(*names), delim="/")

    def prefixed(self, *names: str, delim: SupportedDelim = "-") -> str:
        """Returns as <env_base>-<name1>-<name2>..."""
        # Don't add if self in names already
        if names and (names[0] == self or names[0].startswith(self + delim)):
            return self.concat(*names, delim=delim)
        return self.concat(self, *names, delim=delim)

    def suffixed(self, *names: str, delim: SupportedDelim = "-") -> str:
        """Returns as <name1>-...<nameN>-<env_base>"""
        # Don't add if self in names already
        if names and (names[-1] == self or names[-1].endswith(delim + self)):
            return self.concat(*names, delim=delim)
        return self.concat(*names, self, delim=delim)

    @classmethod
    def concat(cls, *names: str, delim: SupportedDelim = "-") -> str:
        return delim.join(names)

    def to_env(self):
        """Write as environnment variable"""
        set_env_var(ENV_BASE_KEY, self)

    def to_type_and_label(self) -> Tuple[EnvType, Optional[str]]:
        env_type, env_label = self.get_match_groups()
        return (EnvType(env_type), env_label)

    @classmethod
    def from_type_and_label(cls: Type[E], env_type: EnvType, env_label: Optional[str] = None) -> E:
        """Returns <env_name>[-<env_label>]"""
        return cls(f"{env_type.value}-{env_label}" if env_label else env_type.value)

    @classmethod
    def from_env(cls: Type[E]) -> E:
        """Get value from environment variables

        1. Checks for env base variables.
        2. Checks for env type and label:

        Environment Variable Keys:
        env base  : "env_base" / "ENV_BASE"
        env type  : "env_type" / "ENV_TYPE"
        env label : "env_label" / "ENV_LABEL" / "label" / "LABEL"
        Raises:
            ApplicationException: if variables are not found.

        Returns:
            EnvBase
        """
        env_base = get_env_var(ENV_BASE_KEY, ENV_BASE_KEY_ALIAS)
        if env_base:
            return cls(env_base)
        else:
            try:
                env_type = cls.load_env_type__from_env()
                env_label = cls.load_env_label__from_env()
            except:
                raise ApplicationException("Could not resolve Env Base or Env Type from env")

            env_base = cls.from_type_and_label(EnvType(env_type), env_label=env_label)
            return env_base

    @classmethod
    def load_env_type__from_env(cls) -> EnvType:
        env_type = get_env_var(ENV_TYPE_KEY, ENV_TYPE_KEY_ALIAS)
        if not env_type:
            raise ApplicationException(
                f"{cls} is not found as {ENV_TYPE_KEY} or {ENV_TYPE_KEY_ALIAS}"
            )
        return EnvType(env_type)

    @classmethod
    def load_env_label__from_env(cls) -> Optional[str]:
        return get_env_var(ENV_LABEL_KEY, ENV_LABEL_KEY_ALIAS, LABEL_KEY, LABEL_KEY_ALIAS)


class EnvBaseMixinsBase(Generic[E]):
    @property
    def env_base(self) -> E:
        """returns env base

        If env base has not been set, it sets the value using environment variables

        Returns:
            EnvBase: env base
        """
        try:
            return self._env_base
        except AttributeError:
            self.env_base = None  # type: ignore[assignment]
            return self._env_base

    @env_base.setter
    def env_base(self, env_base: Optional[E] = None):
        """Sets env base

        If None is provided, env variables are read to infer env base

        Args:
            env_base (Optional[EnvBase], optional): environment base to use. Defaults to None.
        """
        self._env_base = get_env_base(env_base, env_base_class=self._env_base_class())

    @classmethod
    def _env_base_class(cls) -> Type[E]:
        return cls.__orig_bases__[0].__args__[0]  # type: ignore


# ----------------------------------
#        Mixins & Enums
# ----------------------------------


class EnvBaseMixins(EnvBaseMixinsBase[EnvBase]):
    @classmethod
    def _env_base_class(cls) -> Type[EnvBase]:
        return EnvBase


class EnvBaseEnumMixins:
    def prefix_with(
        self, env_base: Optional[E] = None, env_base_class: Optional[Type[E]] = None
    ) -> str:
        env_base_ = get_env_base(env_base, env_base_class)
        if isinstance(self, Enum):
            return env_base_.prefixed(self.value)
        else:
            return env_base_.prefixed(str(self))


class ResourceNameBaseEnum(str, EnvBaseEnumMixins, Enum):
    def __str__(self) -> str:
        return self.value

    def get_name(self, env_base: EnvBase) -> str:
        return self.prefix_with(env_base)


# ----------------------------------
#        getter functions
# ----------------------------------


@overload
def get_env_base() -> EnvBase:
    ...


@overload
def get_env_base(env_base: Union[str, EnvBase]) -> EnvBase:
    ...


@overload
def get_env_base(env_base: Literal[None]) -> EnvBase:
    ...


@overload
def get_env_base(env_base: Literal[None], env_base_class: Literal[None]) -> EnvBase:
    ...


@overload
def get_env_base(env_base: Literal[None], env_base_class: Type[E]) -> E:
    ...


@overload
def get_env_base(env_base: Union[str, E], env_base_class: Type[E]) -> E:
    ...


@overload
def get_env_base(env_base: Union[str, E], env_base_class: Literal[None]) -> EnvBase:
    ...


def get_env_base(
    env_base: Optional[Union[str, E]] = None,
    env_base_class: Optional[Type[Union[E, EnvBase]]] = None,
) -> Union[E, EnvBase]:
    """Will look for the env_base as an environment variable."""
    env_base_cls: Type[E] = env_base_class or EnvBase  # type: ignore[assignment]
    if env_base:
        if isinstance(env_base, env_base_cls):
            return env_base
        else:
            return env_base_cls(env_base)
    return env_base_cls.from_env()


def get_env_type(
    env_type: Optional[Union[str, EnvType]] = None,
    default_env_type: Optional[EnvType] = None,
    env_base_class: Optional[E] = None,
) -> EnvType:
    """Loads EnvType from environment or normalizes input

    Note that if loading from environment, EnvBase variable takes precedence over EnvType

    """
    if env_type:
        return env_type if isinstance(env_type, EnvType) else EnvType(env_type)
    try:
        return (env_base_class or EnvBase).from_env().env_type
    except Exception as e:
        if default_env_type:
            return default_env_type
        raise e


class _Missing:
    pass


MISSING = _Missing()


def get_env_label(
    env_label: Union[Optional[str], _Missing] = MISSING,
    env_base_class: Optional[E] = None,
) -> Optional[str]:
    """Get Environment label

    If not specified, it will be loaded from envirionment (First checking for env base, then label)

    Args:
        env_label (Union[Optional[str], _Missing], optional): env label. Defaults to MISSING.

    Returns:
        Optional[str]: env label string
    """
    env_base_cls = env_base_class or EnvBase
    if isinstance(env_label, _Missing):
        try:
            # First check if EnvBase exists and
            return env_base_cls.from_env().env_label
        except:
            # next check for env label.
            return env_base_cls.load_env_label__from_env()
    # Right now env label regex is only baked into EnvBase, so let's
    # create an EnvBase to validate
    return env_base_cls.from_type_and_label(EnvType.DEV, env_label=env_label).env_label



import re
from dataclasses import dataclass
from functools import total_ordering
from typing import ClassVar, Optional, Pattern

from aibs_informatics_core.collections import ValidatedStr


@total_ordering
@dataclass
class Version:
    major_version: int
    minor_version: Optional[int] = None
    revision: Optional[int] = None

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Version):
            return (self.major_version, self.minor_version or -1, self.revision or -1) == (
                other.major_version,
                other.minor_version or -1,
                other.revision or -1,
            )
        elif isinstance(other, VersionStr):
            return self == other.version
        else:
            return False

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Version):
            return (self.major_version, self.minor_version or -1, self.revision or -1) < (
                other.major_version,
                other.minor_version or -1,
                other.revision or -1,
            )
        elif isinstance(other, VersionStr):
            return self < other.version
        else:
            return NotImplemented


class VersionStr(ValidatedStr):

    """Version String

    Version String      Major   Minor   Revision
    --------------   ------------------------------
        1.0.0             1       0        0
        1.0.*             1       0       any
        1.0               1       0       any
        1.*               1      any      any
        1                 1      any      any


    """

    regex_pattern: ClassVar[Pattern] = re.compile(r"(?:v?)(?:(\d+))(?:\.(\d+))?(?:\.(\*|\d+))?")

    @property
    def version(self) -> Version:
        major_str, minor_str, revision_str = self.get_match_groups()
        return Version(
            major_version=int(major_str),
            minor_version=int(minor_str) if minor_str and minor_str != "*" else None,
            revision=int(revision_str) if revision_str and revision_str != "*" else None,
        )

    @property
    def major_version(self) -> int:
        return self.version.major_version

    @property
    def minor_version(self) -> Optional[int]:
        return self.version.minor_version

    @property
    def revision(self) -> Optional[int]:
        return self.version.revision

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Version):
            return self.version == other
        elif isinstance(other, VersionStr):
            return self.version == other.version
        elif isinstance(other, str):
            try:
                other = VersionStr(other)
            except ValueError:
                return False
            else:
                return self == other
        else:
            return False

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Version):
            return self.version < other
        elif isinstance(other, VersionStr):
            return self.version < other.version
        else:
            return NotImplemented



import collections
from copy import deepcopy
from typing import Dict, List, Match, Set

from aibs_informatics_core.exceptions import ValidationError
from aibs_informatics_core.models.demand_execution.job_param import JobParam, JobParamRef
from aibs_informatics_core.utils.decorators import cache


class JobParamResolver:
    @classmethod
    @cache
    def find_collisions(cls, *job_params: JobParam) -> List[List[JobParam]]:
        # Validate params do not have conflicting envnames
        envname_job_params: Dict[str, List[JobParam]] = collections.defaultdict(list)
        for job_param in job_params:
            envname_job_params[job_param.envname].append(job_param)
        return [_ for _ in envname_job_params.values() if len(_) > 1]

    @classmethod
    def check_collisions(cls, job_params: List[JobParam]):
        colliding_job_params = cls.find_collisions(*job_params)
        if len(colliding_job_params):
            raise ValidationError(
                "Job Params have at least two parameters with conflicting envnames. "
                f"Complete List of offending job params: {colliding_job_params}"
            )

    @classmethod
    def resolve_references(cls, job_params: List[JobParam]) -> List[JobParam]:
        """Resolves param references found in values of param map

        Valid Cases:

            Example 1: Nested references
                Unresolved: [("x", "${Y}/${Z}_foo"),       ("y", "${Z}_${Z}_bar"), ("z", "qaz")]
                Resolved:   [("x", "qaz_qaz_bar/qaz_foo"), ("y", "qaz_qaz_bar"),   ("z", "qaz")]

            Example 2: Reference by unnormalized env names
                Unresolved: [("reference_a", "foo_${refERENCE_b}"), ("REFerence-B", "bar")]
                Resolved:   [("reference_a", "foo_bar"),            ("REFerence-B", "bar")]

        Invalid Cases:

            Example 1: Cyclical References
                [({"x", "${Y}"), ("y", "${X}")]

            Example 2: Self Reference
                [("X": "${X}")]

            Example 3: Reference to missing param
                [("X", "${Y}")]

        Args:
            job_params (List[JobParam]): List of job params with unresolved references in values

        Returns:
            List[JobParam]: list of job params with resolved references in values
        """

        cls.check_collisions(job_params)

        # job_param.env_name -> job_param (with resolved values)
        resolved_job_param_map: Dict[str, JobParam] = dict()

        # job_param.env_name -> job_param (with unresolved values)
        unresolved_job_param_map: Dict[str, JobParam] = dict()

        # job_param.env_name -> Set of other job_param.env_name references found in job_param.value
        job_param_dependency_map: Dict[str, Set[str]] = dict()

        for job_param in deepcopy(job_params):
            job_param_dependency_map[job_param.envname] = {
                _.envname for _ in job_param.find_references()
            }
            if job_param_dependency_map[job_param.envname]:
                unresolved_job_param_map[job_param.envname] = job_param
            else:
                resolved_job_param_map[job_param.envname] = job_param

        def replace_reference(match: Match) -> str:
            ref = JobParamRef(match.group(0))
            return resolved_job_param_map[JobParam.as_envname(ref.envname)].value

        iter_resolved_count = len(resolved_job_param_map)
        while unresolved_job_param_map:
            for job_param_envname in list(unresolved_job_param_map.keys()):
                job_param_dependencies = job_param_dependency_map[job_param_envname]
                if all([_ in resolved_job_param_map for _ in job_param_dependencies]):
                    job_param = unresolved_job_param_map[job_param_envname]
                    job_param.replace_references(reference_replacement=replace_reference)
                    resolved_job_param_map[job_param_envname] = job_param
                    unresolved_job_param_map.pop(job_param_envname)
            if iter_resolved_count == len(resolved_job_param_map):
                raise ValidationError(
                    "Could not resolve any more references: "
                    f"Resolved: {resolved_job_param_map}, "
                    f"Unresolved: {unresolved_job_param_map}, "
                    f"Reference Dependencies: {job_param_dependency_map}"
                )
        return [resolved_job_param_map[_.envname] for _ in job_params]
